@@ -1,6 +1,6 @@
 from guillotina import directives
 from guillotina.catalog.catalog import DefaultCatalogDataAdapter
-from guillotina.component import getUtilitiesFor
+from guillotina.component import get_utilities_for
 from guillotina.content import iter_schemata_for_type
 from guillotina.db.cache.dummy import DummyCache
 from guillotina.directives import merged_tagged_value_dict
@@ -13,10 +13,13 @@ from guillotina.interfaces import IInteraction
 from guillotina.interfaces import IResourceFactory
 from guillotina.interfaces import ISecurityInfo
 from guillotina.transactions import managed_transaction
+from guillotina.utils import apply_coroutine
 from guillotina.utils import get_content_path
 from guillotina.utils import get_current_request
-from guillotina_elasticsearch.utility import ElasticSearchUtility
 from guillotina_elasticsearch.utils import noop_response
+from guillotina.event import notify
+from guillotina_elasticsearch.events import IndexProgress
+from os.path import join
 
 import aioes
 import asyncio
@@ -24,10 +27,16 @@ import gc
 import json
 import logging
 import resource
-import threading
 import time
 
 
+try:
+    from guillotina.utils import clear_conn_statement_cache
+except ImportError:
+    def clear_conn_statement_cache(conn):
+        pass
+
+
 logger = logging.getLogger('guillotina_elasticsearch')
 
 
@@ -36,11 +45,12 @@ class Indexer:
     def __init__(self):
         self.data_adapter = DefaultCatalogDataAdapter(None)
         self.mappings = {}
-        for type_name, schema in getUtilitiesFor(IResourceFactory):
+        for type_name, schema in get_utilities_for(IResourceFactory):
             self.mappings[type_name] = {}
             for schema in iter_schemata_for_type(type_name):
-                for index_name, index_data in merged_tagged_value_dict(
+                for field_name, index_data in merged_tagged_value_dict(
                         schema, directives.index.key).items():
+                    index_name = index_data.get('index_name', field_name)
                     self.mappings[type_name][index_name] = {
                         'schema': schema,
                         'properties': index_data
@@ -58,51 +68,13 @@ class Indexer:
             await behavior.load(create=False)
         try:
             if 'accessor' in index_data:
-                return index_data['accessor'](behavior)
+                return await apply_coroutine(index_data['accessor'], behavior)
             else:
                 return self.data_adapter.get_data(behavior, schema, index_name)
         except NoIndexField:
             pass
 
 
-class ElasticThread(threading.Thread):
-    def __init__(self, index_name, batch):
-        self.index_name = index_name
-        self.batch = batch
-        super().__init__(target=self)
-
-    def __call__(self):
-        self._loop = asyncio.new_event_loop()
-        self._loop.run_until_complete(self._run())
-
-    async def _run(self):
-        utility = ElasticSearchUtility(loop=self._loop)
-        bulk_data = []
-
-        uids = []
-        for uuid, batch_type, data in self.batch:
-            uids.append(uuid)
-            doc_type = data['type_name']
-            if batch_type == 'update':
-                data = {'doc': data}
-            bulk_data.append({
-                batch_type: {
-                    '_index': self.index_name,
-                    '_type': doc_type,
-                    '_id': uuid
-                }
-            })
-            if batch_type != 'delete':
-                bulk_data.append(data)
-        results = await utility.conn.bulk(index=self.index_name, doc_type=None,
-                                          body=bulk_data)
-        if results['errors']:
-            logger.warn(f'Error bulk bulk putting: {", ".join(uids)}')
-        utility.conn.close()
-        del utility
-        del self.batch
-
-
 def _clean_mapping(mapping):
     if 'properties' in mapping:
         for key in ('confirm',):
@@ -179,16 +151,15 @@ class Migrator:
 
         if request is None:
             self.request = get_current_request()
-            self.request._db_write_enabled = False
         else:
             self.request = request
         # make sure that we don't cache requests...
-        self.request._txn._cache = DummyCache(None, None)
+        self.request._txn._cache = DummyCache(self.request._txn)
         self.container = self.request.container
         self.interaction = IInteraction(self.request)
         self.indexer = Indexer()
 
-        self.batch = []
+        self.batch = {}
         self.indexed = 0
         self.processed = 0
         self.missing = []
@@ -197,7 +168,7 @@ class Migrator:
         self.errors = []
         self.mapping_diff = {}
         self.start_time = self.index_start_time = time.time()
-        self.reindex_threads = []
+        self.reindex_futures = []
         self.status = 'started'
         self.active_task_id = None
 
@@ -228,47 +199,45 @@ class Migrator:
         conn_es = await self.conn.transport.get_connection()
         real_index_name = await self.utility.get_index_name(self.container,
                                                             self.request)
-        resp = await conn_es._session.post(
-            str(conn_es._base_url) + '_reindex',
-            params={
-                'wait_for_completion': 'false'
-            },
-            data=json.dumps({
-              "source": {
-                "index": real_index_name,
-                "size": 100
-              },
-              "dest": {
-                "index": self.work_index_name
-              }
-            })
-        )
-
-        data = await resp.json()
-        self.active_task_id = task_id = data['task']
-        while True:
-            await asyncio.sleep(10)
-            resp = await conn_es._session.get(
-                str(conn_es._base_url) + '_tasks/' + task_id)
-            if resp.status in (400, 404):
-                break
+        async with conn_es._session.post(
+                str(conn_es._base_url) + '_reindex',
+                params={
+                    'wait_for_completion': 'false'
+                },
+                data=json.dumps({
+                    "source": {
+                        "index": real_index_name,
+                        "size": 100
+                    },
+                    "dest": {
+                        "index": self.work_index_name
+                    }
+                })) as resp:
             data = await resp.json()
-            if data['completed']:
-                break
-            status = data["task"]["status"]
-            self.response.write(f'{status["created"]}/{status["total"]} - '
-                                f'Copying data to new index. task id: {task_id}')
-            self.copied_docs = status["created"]
+            self.active_task_id = task_id = data['task']
+            while True:
+                await asyncio.sleep(10)
+                async with conn_es._session.get(
+                        str(conn_es._base_url) + '_tasks/' + task_id) as resp:
+                    if resp.status in (400, 404):
+                        break
+                    data = await resp.json()
+                    if data['completed']:
+                        break
+                    status = data["task"]["status"]
+                    self.response.write(f'{status["created"]}/{status["total"]} - '
+                                        f'Copying data to new index. task id: {task_id}')
+                    self.copied_docs = status["created"]
 
-        self.active_task_id = None
-        response = data['response']
-        failures = response['failures']
-        if len(failures) > 0:
-            failures = json.dumps(failures, sort_keys=True, indent=4,
-                                  separators=(',', ': '))
-            self.response.write(f'Reindex encountered failures: {failures}')
-        else:
-            self.response.write(f'Finished copying to new index: {self.copied_docs}')
+            self.active_task_id = None
+            response = data['response']
+            failures = response['failures']
+            if len(failures) > 0:
+                failures = json.dumps(failures, sort_keys=True, indent=4,
+                                      separators=(',', ': '))
+                self.response.write(f'Reindex encountered failures: {failures}')
+            else:
+                self.response.write(f'Finished copying to new index: {self.copied_docs}')
 
     async def get_all_uids(self):
         self.response.write('Retrieving existing doc ids')
@@ -311,6 +280,7 @@ class Migrator:
         next_mappings = await self.conn.indices.get_mapping(self.work_index_name)
         next_mappings = next_mappings[self.work_index_name]['mappings']
 
+        changes = False
         for type_name in existing_mappings.keys():
             if type_name not in next_mappings:
                 # copy over orphaned type otherwise move will potentially not work
@@ -328,8 +298,13 @@ class Migrator:
                 # and install new mapping
                 await self.utility.conn.indices.put_mapping(
                     self.work_index_name, type_name, mapping)
+                changes = True
+        if changes:
+            # we add to the mappings so we need to update...
+            next_mappings = await self.conn.indices.get_mapping(self.work_index_name)
+            next_mappings = next_mappings[self.work_index_name]['mappings']
 
-        for type_name, schema in getUtilitiesFor(IResourceFactory):
+        for type_name, schema in get_utilities_for(IResourceFactory):
             new_definitions = {}
             if type_name not in existing_mappings:
                 diffs[type_name] = next_mappings[type_name]['properties']
@@ -381,7 +356,7 @@ class Migrator:
             - else, do nothing
             - remove for list of existing doc ids
         '''
-        # do stuff...
+        clear_conn_statement_cache(await ob._p_jar.get_connection())
         full = False
         if ob.uuid not in self.existing:
             self.missing.append(ob.uuid)
@@ -401,9 +376,17 @@ class Migrator:
     async def index_object(self, ob, full=False):
         batch_type = 'update'
         if self.reindex_security:
-            data = ISecurityInfo(ob)()
+            try:
+                data = ISecurityInfo(ob)()
+            except TypeError:
+                self.response.write(f'could not index {ob}')
+                return
         elif full or self.full:
-            data = await ICatalogDataAdapter(ob)()
+            try:
+                data = await ICatalogDataAdapter(ob)()
+            except TypeError:
+                self.response.write(f'could not index {ob}')
+                return
             batch_type = 'index'
         else:
             if ob.type_name not in self.mapping_diff:
@@ -422,7 +405,10 @@ class Migrator:
                     data[index_name] = val
 
         self.indexed += 1
-        self.batch.append((ob.uuid, batch_type, data))
+        self.batch[ob.uuid] = {
+            'action': batch_type,
+            'data': data
+        }
 
         if self.log_details:
             self.response.write(f'({self.processed} {int(self.per_sec())}) '
@@ -439,37 +425,74 @@ class Migrator:
             gc.collect()
             if self.memory_tracking:
                 total_memory = round(
-                    resource.getrusage(resource.RUSAGE_SELF).ru_maxrss/1024.0, 1)
+                    resource.getrusage(resource.RUSAGE_SELF).ru_maxrss / 1024.0, 1)
                 self.response.write(b'Memory usage: % 2.2f MB, cleaned: %d, total in-memory obs: %d' % (
                     total_memory, num, len(gc.get_objects())))
             self.response.write(b'Indexing new batch, totals: (%d %d/sec)\n' % (
                 self.indexed, int(self.per_sec()),
             ))
-
         if len(self.batch) >= self.bulk_size:
+            await notify(IndexProgress(
+                self.request, self.context, self.processed,
+                (len(self.existing) + len(self.missing))
+            ))
             await self.flush()
 
-    def join_threads(self):
-        for thread in self.reindex_threads:
-            thread.join()
-        self.reindex_threads = []
+    async def join_futures(self):
+        for future in self.reindex_futures:
+            if not future.done():
+                await asyncio.wait_for(future, None)
+        self.reindex_futures = []
+
+    async def _index_batch(self, batch):
+        bulk_data = []
+        for _id, payload in batch.items():
+            doc_type = payload['data']['type_name']
+            action_data = {
+                '_index': self.work_index_name,
+                '_type': doc_type,
+                '_id': _id
+            }
+            data = payload['data']
+            if payload['action'] == 'update':
+                data = {'doc': data}
+                action_data['_retry_on_conflict'] = 3
+            bulk_data.append({
+                payload['action']: action_data
+            })
+            if payload['action'] != 'delete':
+                bulk_data.append(data)
+        results = await self.utility.conn.bulk(
+            index=self.work_index_name, doc_type=None,
+            body=bulk_data)
+        if results['errors']:
+            errors = []
+            for result in results['items']:
+                for key, value in result.items():
+                    if not isinstance(value, dict):
+                        continue
+                    if 'status' in value and value['status'] != 200:
+                        _id = value.get('_id')
+                        errors.append(f'{_id}: {value["status"]}')
+
+                        # retry conflict errors and thread pool capacity limits
+                        if value['status'] in (409, 429):
+                            self.batch[_id] = batch[_id]
+            logger.warning(f'Error bulk putting: {results}')
 
     async def flush(self):
         if len(self.batch) == 0:
             # nothing to flush
             return
 
-        thread = ElasticThread(
-            self.work_index_name,
+        future = asyncio.ensure_future(self._index_batch(
             self.batch
-        )
-        self.batch = []
+        ))
+        self.batch = {}
+        self.reindex_futures.append(future)
 
-        self.reindex_threads.append(thread)
-        thread.start()
-
-        if len(self.reindex_threads) > 7:
-            self.join_threads()
+        if len(self.reindex_futures) > 7:
+            await self.join_futures()
 
     async def check_existing(self):
         '''
@@ -488,7 +511,10 @@ class Migrator:
                 try:
                     doc = await self.conn.get(self.work_index_name, uuid,
                                               _source=False)
-                    self.batch.append((uuid, 'delete', {'type_name': doc['_type']}))
+                    self.batch[uuid] = {
+                        'action': 'delete',
+                        'data': {'type_name': doc['_type']}
+                    }
                     await self.attempt_flush()
                     # no longer present on db, this was orphaned
                     self.orphaned.append(uuid)
@@ -506,7 +532,7 @@ class Migrator:
                 # doc = self.utility.conn.get(
                 #     await self.utility.get_index_name(), fields='path'
                 # )
-                # ob = await do_traverse(self.request, self.container,
+                # ob = await traverse(self.request, self.container,
                 #                        doc['_source']['path'].strip('/').split('/'))
                 # await self.index_object(ob, full=True)
 
@@ -531,9 +557,9 @@ class Migrator:
         if self.active_task_id is not None:
             self.response.write('Canceling copy of index task')
             conn_es = await self.conn.transport.get_connection()
-            await conn_es._session.post(
-                str(conn_es._base_url) + '_tasks/' + self.active_task_id + '/_cancel')
-            asyncio.sleep(5)
+            async with conn_es._session.post(
+                    str(conn_es._base_url) + '_tasks/' + self.active_task_id + '/_cancel'):
+                asyncio.sleep(5)
         if self.work_index_name:
             self.response.write('Deleting new index')
             await self.conn.indices.delete(self.work_index_name)
@@ -567,7 +593,7 @@ class Migrator:
             await self.check_existing()
 
             await self.flush()
-            self.join_threads()
+            await self.join_futures()
 
         async with self.utility._migration_lock:
             self.response.write('Activating new index')
@@ -575,7 +601,9 @@ class Migrator:
                 await self.utility.apply_next_index(self.container, self.request)
             self.status = 'done'
 
-            self.response.write('Update aliases')
+            self.response.write(f'''Update alias({alias_index_name}):
+{existing_index} -> {self.work_index_name}
+''')
             await self.conn.indices.update_aliases({
                 "actions": [
                     {"remove": {
