@@ -1,16 +1,16 @@
 # -*- coding: utf-8 -*-
 from guillotina import configure
-from guillotina.async import IAsyncUtility
 from guillotina.event import notify
 from guillotina.interfaces import IAbsoluteURL
 from guillotina.interfaces import ICatalogUtility
 from guillotina.interfaces import IInteraction
-from guillotina.traversal import do_traverse
 from guillotina.utils import get_content_depth
 from guillotina.utils import get_content_path
 from guillotina.utils import get_current_request
 from guillotina.utils import merge_dicts
+from guillotina.utils import navigate_to
 from guillotina_elasticsearch.events import SearchDoneEvent
+from guillotina_elasticsearch.exceptions import QueryErrorException
 from guillotina_elasticsearch.manager import ElasticSearchManager
 from guillotina_elasticsearch.utils import noop_response
 
@@ -19,7 +19,12 @@ import asyncio
 import json
 import logging
 import time
-import uuid
+
+
+try:
+    from guillotina.async_util import IAsyncUtility
+except ImportError:
+    from guillotina.async import IAsyncUtility
 
 
 logger = logging.getLogger('guillotina_elasticsearch')
@@ -39,10 +44,11 @@ class ElasticSearchUtility(ElasticSearchManager):
     index_count = 0
 
     async def reindex_all_content(
-            self, obj, security=False, response=noop_response):
+            self, obj, security=False, response=noop_response,
+            request=None):
         from guillotina_elasticsearch.reindex import Reindexer
         reindexer = Reindexer(self, obj, response=response,
-                              reindex_security=security)
+                              reindex_security=security, request=request)
         await reindexer.reindex(obj)
 
     async def search(self, container, query):
@@ -57,7 +63,8 @@ class ElasticSearchUtility(ElasticSearchManager):
             query,
             doc_type=None,
             size=10,
-            request=None):
+            request=None,
+            scroll=None):
         if query is None:
             query = {}
 
@@ -77,8 +84,7 @@ class ElasticSearchUtility(ElasticSearchManager):
             request = get_current_request()
         interaction = IInteraction(request)
 
-        for user in interaction.participations:
-            print(f'searching with {user.principal.id} {user.principal.groups}')
+        for user in interaction.participations:  # pylint: disable=E1133
             users.append(user.principal.id)
             users.extend(user.principal.groups)
             roles_dict = interaction.global_principal_roles(
@@ -108,12 +114,47 @@ class ElasticSearchUtility(ElasticSearchManager):
         # query.update(permission_query)
         q['body'] = query
         q['size'] = size
-        logger.warn(q)
+
+        if scroll:
+            q['scroll'] = scroll
+
+        logger.debug(q)
         return q
 
+    def _get_items_from_result(self, container, request, result):
+        items = []
+        container_url = IAbsoluteURL(container, request)()
+        for item in result['hits']['hits']:
+            data = item.pop('_source', {})
+            for key, val in item.get('fields', {}).items():
+                container_data = data
+                if isinstance(val, list):
+                    if len(val) == 1:
+                        val = val[0]
+                    else:
+                        val = None
+                if '.' in key:
+                    name, key = key.split('.', 1)
+                    if name not in container_data:
+                        container_data[name] = {}
+                    container_data = container_data[name]
+                container_data[key] = val
+            data.update({
+                '@absolute_url': container_url + data.get('path', ''),
+                '@type': data.get('type_name'),
+                '@uid': item['_id'],
+                '@name': data.get('id', data.get('path', '').split('/')[-1])
+            })
+            sort_value = item.get('sort')
+            if sort_value:
+                data.update({'sort': sort_value})
+            items.append(data)
+        return items
+
     async def query(
             self, container, query,
-            doc_type=None, size=10, request=None):
+            doc_type=None, size=10, request=None,
+            scroll=None):
         """
         transform into query...
         right now, it's just passing through into elasticsearch
@@ -122,20 +163,15 @@ class ElasticSearchUtility(ElasticSearchManager):
         if request is None:
             request = get_current_request()
         q = await self._build_security_query(
-            container, query, doc_type, size, request)
+            container, query, doc_type, size, request, scroll)
         result = await self.conn.search(**q)
-        items = []
-        container_url = IAbsoluteURL(container, request)()
-        for item in result['hits']['hits']:
-            data = item['_source']
-            data.update({
-                '@absolute_url': container_url + data.get('path', ''),
-                '@type': data.get('type_name'),
-            })
-            sort_value = item.get('sort')
-            if sort_value:
-                data.update({'sort': sort_value})
-            items.append(data)
+        if result.get('_shards', {}).get('failed', 0) > 0:
+            logger.warning(f'Error running query: {result["_shards"]}')
+            error_message = 'Unknown'
+            for failure in result["_shards"].get('failures') or []:
+                error_message = failure['reason']
+            raise QueryErrorException(reason=error_message)
+        items = self._get_items_from_result(container, request, result)
         final = {
             'items_count': result['hits']['total'],
             'member': items
@@ -144,8 +180,13 @@ class ElasticSearchUtility(ElasticSearchManager):
             final['aggregations'] = result['aggregations']
         if 'suggest' in result:
             final['suggest'] = result['suggest']
+        if 'profile' in result:
+            final['profile'] = result['profile']
+        if '_scroll_id' in result:
+            final['_scroll_id'] = result['_scroll_id']
+
         tdif = time.time() - t1
-        print('Time ELASTIC %f' % tdif)
+        logger.debug(f'Time ELASTIC {tdif}')
         await notify(SearchDoneEvent(
             query, result['hits']['total'], request, tdif))
         return final
@@ -179,15 +220,20 @@ class ElasticSearchUtility(ElasticSearchManager):
             raise AttributeError('Not found a unique object')
 
         path = result['members'][0]['path']
-        obj = do_traverse(container, path)
+        obj = await navigate_to(container, path)
         return obj
 
-    async def get_by_type(self, container, doc_type, query={}, size=10):
+    async def get_by_type(self, container, doc_type, query=None, size=10):
+        if query is None:
+            query = {}
         return await self.query(container, query, doc_type=doc_type, size=size)
 
     async def get_by_path(
-            self, container, path, depth=-1, query={}, doc_type=None, size=10):
-        if type(path) is not str:
+            self, container, path, depth=-1, query=None,
+            doc_type=None, size=10, scroll=None):
+        if query is None:
+            query = {}
+        if not isinstance(path, str):
             path = get_content_path(path)
 
         if path is not None and path != '/':
@@ -195,7 +241,7 @@ class ElasticSearchUtility(ElasticSearchManager):
                 'query': {
                     'bool': {
                         'must': [{
-                            'match': {'path': path}
+                            'term': {'path': path}
                         }]
                     }
                 }
@@ -208,23 +254,23 @@ class ElasticSearchUtility(ElasticSearchManager):
             query = merge_dicts(query, path_query)
             # We need the local roles
 
-        return await self.query(container, query, doc_type, size=size)
+        return await self.query(container, query, doc_type,
+                                size=size, scroll=scroll)
 
-    async def call_unindex_all_childs(self, index_name, path_query):
+    async def call_unindex_all_children(self, index_name, path_query):
         conn_es = await self.conn.transport.get_connection()
         async with conn_es._session.post(
-                    conn_es._base_url.human_repr() + index_name + '/_delete_by_query',
-                    data=json.dumps(path_query)
-                ) as resp:
+                conn_es._base_url.human_repr() + index_name + '/_delete_by_query',
+                data=json.dumps(path_query)) as resp:
             result = await resp.json()
             if 'deleted' in result:
-                logger.warn('Deleted %d childs' % result['deleted'])
-                logger.warn('Deleted %s ' % json.dumps(path_query))
+                logger.debug(f'Deleted {result["deleted"]} children')
+                logger.debug(f'Deleted {json.dumps(path_query)}')
             else:
-                logger.warn('Wrong deletion of childs' + json.dumps(result))
+                self.log_result(result, 'Deletion of children')
 
     async def get_path_query(self, resource, index_name=None, response=noop_response):
-        if type(resource) is str:
+        if isinstance(resource, str):
             path = resource
             depth = path.count('/') + 1
         else:
@@ -257,49 +303,39 @@ class ElasticSearchUtility(ElasticSearchManager):
             })
         return path_query
 
-    async def unindex_all_childs(self, resource, index_name=None,
-                                 response=noop_response, future=True):
+    async def unindex_all_children(self, resource, index_name=None,
+                                   response=noop_response):
         path_query = await self.get_path_query(resource, index_name, response)
-        request = get_current_request()
-        if future:
-            _id = 'unindex_all_childs-' + uuid.uuid4().hex
-            request._futures.update({
-                _id: self.call_unindex_all_childs(index_name, path_query)})
-        else:
-            await self.call_unindex_all_childs(index_name, path_query)
+        await self.call_unindex_all_children(index_name, path_query)
 
-    async def update_by_query(self, query, future=True):
+    async def update_by_query(self, query):
         request = get_current_request()
         index_name = await self.get_index_name(request.container)
         resp = None
-        if future:
-            _id = 'update_by_query-' + uuid.uuid4().hex
-            request._futures.update({
-                _id: self._update_by_query(query, index_name)})
-        else:
-            resp = await self._update_by_query(query, index_name)
+        resp = await self._update_by_query(query, index_name)
 
         next_index_name = await self.get_next_index_name(
             request.container, request=request)
         if next_index_name:
             async with self._migration_lock:
-                _id = 'update_by_path-' + uuid.uuid4().hex
-                request._futures.update({
-                    _id: self._update_by_query(query, next_index_name)})
+                await self._update_by_query(query, next_index_name)
         return resp
 
     async def _update_by_query(self, query, index_name):
         conn_es = await self.conn.transport.get_connection()
-        resp = await conn_es._session.post(
-            conn_es._base_url.human_repr() + index_name + '/_update_by_query?conflicts=proceed',
-            data=json.dumps(query))
-        result = await resp.json()
-        if 'updated' in result:
-            logger.warn('Updated %d childs' % result['updated'])
-            logger.warn('Updated %s ' % json.dumps(query))
-        else:
-            logger.warn('Wrong update of children' + json.dumps(result))
-        return result
+        url = '{}{}/_update_by_query?conflicts=proceed'.format(
+            conn_es._base_url.human_repr(), index_name
+        )
+        async with conn_es._session.post(
+                url, data=json.dumps(query)) as resp:
+            result = await resp.json()
+            if 'updated' in result:
+                logger.debug(f'Updated {result["updated"]} children')
+                logger.debug(f'Updated {json.dumps(query)} ')
+            else:
+                self.log_result(result, 'Updating children')
+
+            return result
 
     async def get_folder_contents(self, container, parent_uuid, doc_type=None):
         query = {
@@ -332,7 +368,7 @@ class ElasticSearchUtility(ElasticSearchManager):
                 response.write(b'Could not index %s\n' % str(e).encode('utf-8'))
                 logger.error('Could not index ' + ' '.join(idents) + ' ' + str(e))
             else:
-                await asyncio.sleep(1.0)
+                await asyncio.sleep(0.5)
                 result = await self.bulk_insert(index_name, bulk_data, idents, count)
         except aiohttp.client_exceptions.ClientOSError as e:
             count += 1
@@ -340,7 +376,7 @@ class ElasticSearchUtility(ElasticSearchManager):
                 response.write(b'Could not index %s\n' % str(e).encode('utf-8'))
                 logger.error('Could not index ' + ' '.join(idents) + ' ' + str(e))
             else:
-                await asyncio.sleep(1.0)
+                await asyncio.sleep(0.5)
                 result = await self.bulk_insert(index_name, bulk_data, idents, count)
 
         return result
@@ -377,8 +413,8 @@ class ElasticSearchUtility(ElasticSearchManager):
         if len(bulk_data) > 0:
             result = await self.bulk_insert(
                 index_name, bulk_data, idents, response=response)
-        if 'errors' in result and result['errors']:
-            logger.error(json.dumps(result))
+
+        self.log_result(result)
 
         if check_next:
             # also need to call on next index while it's running...
@@ -406,7 +442,8 @@ class ElasticSearchUtility(ElasticSearchManager):
                     'update': {
                         '_index': index_name,
                         '_type': data['type_name'],
-                        '_id': ident
+                        '_id': ident,
+                        '_retry_on_conflict': 3
                     }
                 }, {'doc': data}])
                 idents.append(ident)
@@ -419,14 +456,25 @@ class ElasticSearchUtility(ElasticSearchManager):
             if len(bulk_data) > 0:
                 result = await self.bulk_insert(
                     index_name, bulk_data, idents, response=response)
-            if 'errors' in result and result['errors']:
-                logger.error(json.dumps(result))
+            self.log_result(result)
             return result
 
-    async def remove(self, container, uids, index_name=None, request=None, future=True):
+    def log_result(self, result, label='ES Query'):
+        if 'errors' in result and result['errors']:
+            try:
+                if result['error']['caused_by']['type'] in ('index_not_found_exception',
+                                                            'cluster_block_exception'):
+                    return  # ignore these...
+            except KeyError:
+                return
+            logger.error(label + ': ' + json.dumps(result))
+        else:
+            logger.debug(label + ': ' + json.dumps(result))
+
+    async def remove(self, container, uids, index_name=None, request=None):
         """List of UIDs to remove from index.
 
-        It will remove all the childs on the index"""
+        It will remove all the children on the index"""
         if not self.enabled:
             return
 
@@ -445,8 +493,7 @@ class ElasticSearchUtility(ElasticSearchManager):
                         '_type': type_name
                     }
                 })
-                await self.unindex_all_childs(content_path, index_name=index_name,
-                                              future=future)
+                await self.unindex_all_children(content_path, index_name=index_name)
             await self.conn.bulk(index=index_name, body=bulk_data)
 
         if check_next:
@@ -456,15 +503,15 @@ class ElasticSearchUtility(ElasticSearchManager):
             if next_index_name:
                 async with self._migration_lock:
                     await self.remove(container, uids, next_index_name,
-                                      request=request, future=future)
+                                      request=request)
 
     async def get_doc_count(self, container, index_name=None):
         if index_name is None:
-            index_name = await self.get_index_name(container)
+            index_name = await self.get_real_index_name(container)
         result = await self.conn.count(index=index_name)
         return result['count']
 
     async def refresh(self, container, index_name=None):
         if index_name is None:
-            index_name = await self.get_index_name(container)
+            index_name = await self.get_real_index_name(container)
         await self.conn.indices.refresh(index=index_name)
